import numpy as np
from spglib import get_symmetry_dataset

from spinspg.permutation import get_symmetry_permutations


def test_symmetry_permutations(fcc):
    lattice, positions, numbers, _ = fcc
    symprec = 1e-5
    dataset = get_symmetry_dataset((lattice, positions, numbers), symprec)
    rotations = dataset["rotations"]
    translations = dataset["translations"]
    permutations = get_symmetry_permutations(
        lattice, positions, numbers, rotations, translations, symprec
    )

    assert len(permutations) == len(rotations)
    for permutation in permutations:
        assert np.all(np.sort(permutation.permutation) == np.arange(len(positions)))



from __future__ import annotations

import numpy as np
import pytest
from spglib import get_pointgroup, get_symmetry_from_database
from spgrep.pointgroup import pg_dataset
from spgrep.utils import is_integer_array

from spinspg.pointgroup import (
    POINT_GROUP_GENERATORS,
    POINT_GROUP_REPRESENTATIVES,
    SPIN_POINT_GROUP_TYPES,
    get_integer_point_group,
    get_pointgroup_representative,
    get_pointgroup_representative_from_symbol,
    traverse_spin_operations,
)


def test_point_group_representatives():
    assert set(POINT_GROUP_REPRESENTATIVES.keys()) == set(pg_dataset.keys())


def test_point_group_generators():
    identity = ((1, 0, 0), (0, 1, 0), (0, 0, 1))
    for symbol, generators in POINT_GROUP_GENERATORS.items():
        group = get_pointgroup_representative_from_symbol(symbol)
        generated = traverse_spin_operations(
            [(identity, g) for g in [group[i] for i in generators]]
        )
        assert len(generated) == len(group)


def test_spin_point_group_table():
    founds = set()
    for symbol_R, datum_R in SPIN_POINT_GROUP_TYPES.items():
        R = get_pointgroup_representative_from_symbol(symbol_R)
        generator_R = [R[i] for i in POINT_GROUP_GENERATORS[symbol_R]]
        for symbol_r, datum_R_r in datum_R.items():  # type: ignore
            for symbol_B, datum_R_r_B in datum_R_r.items():
                B = get_pointgroup_representative_from_symbol(symbol_B)
                for number, mapping in datum_R_r_B:
                    print(f"#{number}: R={symbol_R}, r={symbol_r}, B={symbol_B}")
                    spg_generators = [(B[idx], generator_R[i]) for i, idx in enumerate(mapping)]
                    spg = traverse_spin_operations(spg_generators)

                    R2 = list({rot for _, rot in spg})
                    B2 = list({srot for srot, _ in spg})
                    r2 = [rot for srot, rot in spg if srot == ((1, 0, 0), (0, 1, 0), (0, 0, 1))]
                    symbol_r2, _, _ = get_pointgroup(r2)

                    assert len(spg) == len(R)
                    assert set(R2) == set(R)
                    assert set(B2) == set(B)
                    assert symbol_r2 == symbol_r
                    assert spg not in founds
                    founds.add(spg)

    assert len(founds) == 598


@pytest.mark.parametrize(
    "symbol,idx",
    [
        ("mm2", 0),  # unique axis a -> unique axis c
        ("mm2", 1),  # unique axis b -> unique axis c
        ("mm2", 2),  # unique axis c -> unique axis c
        ("-42m", 0),  # -42m -> -42m
        ("-42m", 1),  # -4m2 -> -42m
        ("32", 0),  # 312 -> 312
        ("32", 1),  # 321 -> 312
        ("3m", 0),  # 3m1 -> 3m1
        ("3m", 1),  # 31m -> 3m1
        ("-3m", 0),  # -31m -> -31m
        ("-3m", 1),  # -3m1 -> -31m
        ("-6m2", 0),  # -6m2 -> -6m2
        ("-6m2", 1),  # -62m -> -6m2
    ],
)
def test_get_pointgroup_representative(symbol, idx):
    pg = np.array(pg_dataset[symbol][idx])[::-1]
    symbol_actual, P, mapping = get_pointgroup_representative(pg)

    assert symbol_actual == symbol

    idx_std = POINT_GROUP_REPRESENTATIVES[symbol]
    pg_std = np.array(pg_dataset[symbol][idx_std])
    assert len(pg_std) == len(pg)

    for i in range(len(pg)):
        ri_actual = np.linalg.inv(P) @ pg[mapping[i]] @ P
        assert is_integer_array(ri_actual)
        ri_actual = np.around(ri_actual).astype(int)
        assert np.allclose(ri_actual, pg_std[i])


def test_get_integer_point_group_3m():
    # 3m
    cart_rotations = np.array(
        [
            np.eye(3),
            [
                [np.cos(np.pi * 2 / 3), -np.sin(np.pi * 2 / 3), 0],
                [np.sin(np.pi * 2 / 3), np.cos(np.pi * 2 / 3), 0],
                [0, 0, 1],
            ],
            [
                [np.cos(np.pi * 4 / 3), -np.sin(np.pi * 4 / 3), 0],
                [np.sin(np.pi * 4 / 3), np.cos(np.pi * 4 / 3), 0],
                [0, 0, 1],
            ],
            np.diag([1, -1, 1]),
            [
                [np.cos(np.pi * 2 / 3), -np.sin(np.pi * 2 / 3), 0],
                [-np.sin(np.pi * 2 / 3), -np.cos(np.pi * 2 / 3), 0],
                [0, 0, 1],
            ],
            [
                [np.cos(np.pi * 4 / 3), -np.sin(np.pi * 4 / 3), 0],
                [-np.sin(np.pi * 4 / 3), -np.cos(np.pi * 4 / 3), 0],
                [0, 0, 1],
            ],
        ],
        dtype=np.float_,
    )
    P, rotations = get_integer_point_group(cart_rotations)

    for rot, cart_rot in zip(rotations, cart_rotations):
        rot2 = np.linalg.inv(P) @ cart_rot @ P
        assert is_integer_array(rot2)
        assert np.allclose(rot, rot2)


@pytest.mark.parametrize(
    "hall_number,pointgroup_international",
    [
        (1, "1"),
        (2, "-1"),
        (3, "2"),
        (18, "m"),
        (57, "2/m"),
        (108, "222"),
        (125, "mm2"),
        (227, "mmm"),
        (349, "4"),
        (355, "-4"),
        (357, "4/m"),
        (366, "422"),
        (376, "4mm"),
        (388, "-42m"),
        (400, "4/mmm"),
        (430, "3"),
        (435, "-3"),
        (438, "32"),
        (446, "3m"),
        (454, "-3m"),
        (462, "6"),
        (468, "-6"),
        (469, "6/m"),
        (471, "622"),
        (477, "6mm"),
        (481, "-6m2"),
        (485, "6/mmm"),
        (489, "23"),
        (494, "m-3"),
        (503, "432"),
        (511, "-43m"),
        (517, "m-3m"),
    ],
)
def test_get_integer_point_group(hall_number, pointgroup_international):
    symmetry = get_symmetry_from_database(hall_number)
    rotations_expect = symmetry["rotations"]

    # Generate random orthogonal matrix
    np.random.seed(0)
    P, _ = np.linalg.qr(np.random.rand(3, 3))

    # A hexagonal lattice
    P_hex = np.array(
        [
            [1, 0, 0],
            [-0.5, np.sqrt(3) / 2, 0],
            [0, 0, 1],
        ]
    ).T

    cart_rotations = []
    for rot in rotations_expect:
        if pointgroup_international in [
            "3",
            "-3",
            "32",
            "3m",
            "-3m",
            "6",
            "-6",
            "6/m",
            "622",
            "6mm",
            "-6m2",
            "6/mmm",
        ]:
            rot = P_hex @ rot @ np.linalg.inv(P_hex)

        cart_rot = P @ rot @ np.linalg.inv(P)
        assert np.allclose(cart_rot @ cart_rot.T, np.eye(3))
        cart_rotations.append(cart_rot)

    P_actual, rotations_actual = get_integer_point_group(cart_rotations)

    for rot, cart_rot in zip(rotations_actual, cart_rotations):
        rot2 = np.linalg.inv(P_actual) @ cart_rot @ P_actual
        assert is_integer_array(rot2)
        assert np.allclose(rot, rot2)



"""Utility functions."""
from __future__ import annotations

from typing import Any

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeAlias  # for Python<3.10

NDArrayInt: TypeAlias = NDArray[np.int_]
NDArrayFloat: TypeAlias = NDArray[np.float_]


def ndarray2d_to_integer_tuple(array: NDArrayFloat) -> tuple[tuple[Any]]:
    """Convert two-dimensional array to tuple of tuple."""
    array_int = np.around(array).astype(int)
    array_t = tuple(map(tuple, array_int.tolist()))
    return array_t  # type: ignore


def is_integer_array(array: NDArrayFloat, rtol: float = 1e-5, atol: float = 1e-8) -> bool:
    """Return true if all values of ``array`` are almost integers."""
    array_int = np.around(array).astype(int)
    return np.allclose(array_int, array, rtol=rtol, atol=atol)



"""Permutations from action of symmetry operation on sites."""
from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from spinspg.utils import NDArrayFloat, NDArrayInt


@dataclass
class Permutation:
    """Permutation of list."""

    permutation: NDArrayInt

    def __call__(self, idx: int) -> int:
        """Return a permuted index of ``idx``."""
        return self.permutation[idx]

    def __mul__(self, rhs: Permutation) -> Permutation:
        """Return product with a given permutation ``rhs``.

        (self * rhs)(i) = self(rhs(i))
        """
        n = len(self.permutation)
        assert len(rhs.permutation) == n
        mul = np.zeros(n, dtype=np.int_)
        for i in range(n):
            mul[i] = self(rhs(i))
        return Permutation(mul)


def get_symmetry_permutations(
    lattice: NDArrayFloat,
    positions: NDArrayFloat,
    numbers: NDArrayInt,
    rotations: NDArrayInt,
    translations: NDArrayFloat,
    symprec: float,
) -> list[Permutation]:
    """Return permutations of sites from given symmetry operations."""
    num_sites = len(positions)

    permutations = []
    for rot, trans in zip(rotations, translations):
        new_positions = positions @ rot.T + trans[None, :]
        perm = [-1 for _ in range(num_sites)]
        found = [False for _ in range(num_sites)]
        for i in range(num_sites):
            for j in range(num_sites):
                if found[j] or (numbers[i] != numbers[j]):
                    continue
                if is_overlap_with_origin(lattice, new_positions[i] - positions[j], symprec):
                    perm[i] = j
                    found[j] = True
                    break

        if np.all(perm != -1):
            permutations.append(Permutation(perm))

    return permutations


def is_overlap_with_origin(lattice, frac_coords, symprec) -> bool:
    """Return true iff ``frac_coords`` is overlapped with the origin up to lattice translations."""
    diff = lattice.T @ (frac_coords - np.rint(frac_coords))
    return np.linalg.norm(diff) < symprec



"""Core APIs."""
from __future__ import annotations

import numpy as np

from spinspg.group import get_primitive_spin_symmetry, get_symmetry_with_cell
from spinspg.spin import SpinOnlyGroup
from spinspg.utils import NDArrayFloat, NDArrayInt


def get_spin_symmetry(
    lattice: NDArrayFloat,
    positions: NDArrayFloat,
    numbers: NDArrayInt,
    magmoms: NDArrayFloat,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
) -> tuple[SpinOnlyGroup, NDArrayInt, NDArrayFloat, NDArrayFloat]:
    """Return spin symmetry operations of a given spin arrangement.

    See :ref:`Spglib's document <spglib:py_variables_crystal_structure>` for how to specify the spin arrangement by ``lattice``, ``positions``, ``numbers``, and ``magmoms`` in details.

    Parameters
    ----------
    lattice: array, (3, 3)
        ``lattice[i, :]`` is the ``i``-th basis vector of a lattice
    positions: array, (num_sites, 3)
        ``positions[i, :]`` is a fractional coordinates of the ``i``-th site w.r.t. ``lattice``.
    numbers: array[int], (num_sites, )
        ``numbers[i]`` specifies a specie at the ``i``-th site.
    magmoms: array, (num_sites, 3)
        ``magmoms[i, :]`` is a magnetic moments at the ``i``-th site in Cartesian coordinates.
    symprec: float, default=1e-5
        See :ref:`spglib:variables_symprec`.
    angle_tolerance: float, default=-1
        See :ref:`spglib:variables_angle_tolerance`.

    Returns
    -------
    spin_only_group: :class:`spin.SpinOnlyGroup`
    rotations: array[int], (num_sym, 3, 3)
        Rotation parts of spin symmetry operations w.r.t. ``lattice``.
    translations: array, (num_sym, 3)
        Translation parts of spin symmetry operations w.r.t. ``lattice``.
    spin_rotations: array, (num_sym, 3, 3)
        Spin rotation parts of spin symmetry operations in Cartesian coordinates.
    """
    ns = get_symmetry_with_cell(lattice, positions, numbers, symprec, angle_tolerance)
    ssg = get_primitive_spin_symmetry(ns, magmoms, symprec)

    spin_only_group = ssg.spin_only_group
    tmat = ssg.transformation
    invtmat = np.linalg.inv(tmat)
    rotations = []
    translations = []
    spin_rotations = []

    # Products of "translations in cell", "nontrivial spin translation group's coset", and "nontrivial spin space group's coset"
    for ops in ssg.nontrivial_coset:
        # Transform to primitive to input cell
        new_rotation = np.around(invtmat @ ops.rotation @ tmat).astype(np.int_)
        for ops_st in ssg.spin_translation_coset:
            for centering in ssg.prim_centerings:
                new_translation = np.remainder(
                    invtmat @ (ops.translation + ops_st.translation + centering), 1
                )
                rotations.append(new_rotation)
                translations.append(new_translation)
                spin_rotations.append(ops_st.spin_rotation @ ops.spin_rotation)

    return spin_only_group, np.array(rotations), np.array(translations), np.array(spin_rotations)



"""Spin rotation for magnetic moments."""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

import numpy as np
from spgrep.spinor import get_rotation_angle_and_axis

from spinspg.utils import NDArrayFloat


class SpinOnlyGroupType(Enum):
    """Type of spin only group."""

    NONMAGNETIC = auto()  # O(3)
    COLLINEAR = auto()  # inf/m
    COPLANAR = auto()  # m
    NONCOPLANAR = auto()  # 1

    def __str__(self) -> str:
        """Return string representation."""
        return self.name


@dataclass
class SpinOnlyGroup:
    """Spin only group.

    Attributes
    ----------
    spin_only_group_type: :class:`spin.SpinOnlyGroupType`
        Nonmagnetic, collinear, coplanar, or non-coplanar
    axis: array or None, (3, )
        For collinear and coplanar, unit vector perpendicular to its rotation axis.
    """

    spin_only_group_type: SpinOnlyGroupType
    axis: NDArrayFloat | None

    def __str__(self) -> str:
        """Return string representation."""
        if self.spin_only_group_type in [SpinOnlyGroupType.COLLINEAR, SpinOnlyGroupType.COPLANAR]:
            return str(self.spin_only_group_type) + f"(axis={self.axis})"
        else:
            return str(self.spin_only_group_type)

    def contain(self, linear: NDArrayFloat, atol: float = 1e-5) -> bool:
        """Return if this spin only group contains ``linear``."""
        if self.spin_only_group_type == SpinOnlyGroupType.NONMAGNETIC:
            return True
        if np.allclose(linear, np.eye(3), atol=atol):
            # Group contains identity
            return True

        if np.linalg.det(linear) > 0:
            sign = 1
            rotation = np.array(linear).copy()
        else:
            sign = -1
            rotation = -np.array(linear)

        theta, axis = get_rotation_angle_and_axis(rotation)

        is_parallel = (self.axis is not None) and np.allclose(
            np.cross(axis, self.axis), 0, atol=atol
        )
        two_fold = np.isclose(theta, np.pi, atol=atol)
        if self.spin_only_group_type == SpinOnlyGroupType.COPLANAR:
            if (sign == -1) and is_parallel and two_fold:
                # ``linear`` is a mirror along self.axis
                return True
            else:
                return False
        elif self.spin_only_group_type == SpinOnlyGroupType.COLLINEAR:
            if (sign == 1) and is_parallel:
                # ``linear`` is a rotation along self.axis
                return True
            elif (sign == -1) and np.isclose(np.inner(axis, self.axis), 0, atol=atol) and two_fold:
                return True
            else:
                return False
        else:
            return False

    @classmethod
    def nonmagnetic(cls) -> SpinOnlyGroup:
        """Return nonmagnetic spin-only group."""
        return SpinOnlyGroup(SpinOnlyGroupType.NONMAGNETIC, None)

    @classmethod
    def collinear(cls, axis: NDArrayFloat) -> SpinOnlyGroup:
        """Return collinear spin-only group with the parallel axis."""
        return SpinOnlyGroup(SpinOnlyGroupType.COLLINEAR, axis)

    @classmethod
    def coplanar(cls, axis: NDArrayFloat) -> SpinOnlyGroup:
        """Return coplanar spin-only group with the perpendicular axis."""
        return SpinOnlyGroup(SpinOnlyGroupType.COPLANAR, axis)

    @classmethod
    def noncoplanar(cls) -> SpinOnlyGroup:
        """Return noncoplanar spin-only group."""
        return SpinOnlyGroup(SpinOnlyGroupType.NONCOPLANAR, None)


def get_spin_only_group(magmoms: NDArrayFloat, mag_symprec: float) -> SpinOnlyGroup:
    """Determine spin only group of given spin arrangement.

    Parameters
    ----------
    magmoms : array, (num_sites, 3)
        Magnetic moments in Cartesian coordinates

    Returns
    -------
    spin_only_group: SpinOnlyGroup
    """
    # Nonmagnetic
    if np.max(np.linalg.norm(magmoms, axis=1)) < mag_symprec:
        return SpinOnlyGroup.nonmagnetic()

    moment = np.einsum("ij,ik->jk", magmoms, magmoms, optimize="greedy")  # (3, 3), symmetric
    _, eigvecs = np.linalg.eigh(moment)  # eigenvalues in ascending order

    # Collinear
    parallel_axis = eigvecs[:, -1] / np.linalg.norm(eigvecs[:, -1])
    residual_collinear = (
        magmoms - (magmoms @ parallel_axis)[:, None] * parallel_axis[None, :]
    )  # (N, 3)
    if np.max(2 * np.linalg.norm(residual_collinear, axis=1)) < mag_symprec:
        return SpinOnlyGroup.collinear(parallel_axis)

    # Coplanar
    vertical_axis = eigvecs[:, 0] / np.linalg.norm(eigvecs[:, 0])
    residual_coplanar = (magmoms @ vertical_axis)[:, None] * vertical_axis[None, :]  # (N, 3)
    if np.max(2 * np.linalg.norm(residual_coplanar, axis=1)) < mag_symprec:
        return SpinOnlyGroup.coplanar(vertical_axis)

    # Non-coplanar
    return SpinOnlyGroup.noncoplanar()


def solve_procrustes(A: NDArrayFloat, B: NDArrayFloat) -> NDArrayFloat:
    """Solve orthogonal Procrustes problem.

        argmin_{ R in O(3) } || R A^T - B^T ||_{F}

    Parameters
    ----------
    A: array, (n, 3)
    B: array, (n, 3)

    Returns
    -------
    R: array, (3, 3)
        orthogonal matrix
    """
    u, s, vt = np.linalg.svd(B.T @ A)
    R = u @ vt
    return R



import json
from pathlib import Path

import numpy as np
from spgrep.utils import is_integer_array, ndarray2d_to_integer_tuple

from spinspg.pointgroup import (
    POINT_GROUP_REPRESENTATIVES,
    get_pointgroup_representative_from_symbol,
)


def get_automorphism_permutation(group, g):
    conjugated_group = []
    for h in group:
        conj = np.linalg.inv(g) @ h @ g
        assert is_integer_array(conj)
        conjugated_group.append(ndarray2d_to_integer_tuple(conj))

    assert set(conjugated_group) == set(group)
    mapping = [-1 for _ in range(len(group))]
    for i, hi in enumerate(group):
        j = conjugated_group.index(hi)  # type: ignore
        mapping[j] = i

    return mapping


AUT_D4h = np.array(
    [
        # x, y, z
        [
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ],
        # -y, x, z
        [[0, -1, 0], [1, 0, 0], [0, 0, 1]],
        # -x, y, -z
        [
            [-1, 0, 0],
            [0, 1, 0],
            [0, 0, -1],
        ],
        # y, x, -z
        [
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, -1],
        ],
    ]
)

AUT_D6h = np.array(
    [
        # x, y, z
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        # -y, x-y, z
        [
            [0, -1, 0],
            [1, -1, 0],
            [0, 0, 1],
        ],
        # -x+y, -x, z
        [
            [-1, 1, 0],
            [-1, 0, 0],
            [0, 0, 1],
        ],
        # y, x, -z
        [
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, -1],
        ],
        # -x+y, y, -z
        [
            [-1, 1, 0],
            [0, 1, 0],
            [0, 0, -1],
        ],
        # x, x-y, -z
        [
            [1, 0, 0],
            [1, -1, 0],
            [0, 0, -1],
        ],
    ]
)


def main():
    symbols = POINT_GROUP_REPRESENTATIVES.keys()
    all_datum = {}
    for symbol in symbols:
        print(symbol)
        if symbol in ["222", "mmm"]:
            # Aut:
            # Aut: | m-3m / mmm | = 6
            aut = np.array(
                [
                    # x, y, z
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                    # z, x, y
                    [
                        [0, 0, 1],
                        [1, 0, 0],
                        [0, 1, 0],
                    ],
                    # y, z, x
                    [
                        [0, 1, 0],
                        [0, 0, 1],
                        [1, 0, 0],
                    ],
                    # y, x, -z
                    [
                        [0, 1, 0],
                        [1, 0, 0],
                        [0, 0, -1],
                    ],
                    # x, z, -y
                    [
                        [1, 0, 0],
                        [0, 0, 1],
                        [0, -1, 0],
                    ],
                    # z, y, -x
                    [[0, 0, 1], [0, 1, 0], [-1, 0, 0]],
                ]
            )
        elif symbol in ["23", "m-3", "432", "-43m", "m-3m"]:
            # Aut: 432
            aut = np.array(get_pointgroup_representative_from_symbol("432"))
        elif symbol == "mm2":
            # Aut: | 4/mmm / mmm | = 2
            aut = np.array(
                [
                    [[1, 0, 0], [0, 1, 0], [0, 0, 1]],  # 1
                    [
                        [0, -1, 0],
                        [1, 0, 0],
                        [0, 0, 1],
                    ],  # 4^+
                ]
            )
        elif symbol == "-42m":
            # Aut: | 4/mmm / 2/m | = 4
            aut = AUT_D4h
        elif symbol in ["422", "4mm", "4/mmm"]:
            # Aut: | 8/mmm / 2/m | = 8
            C8 = np.array(
                [
                    [np.cos(np.pi / 4), -np.sin(np.pi / 4), 0],
                    [np.sin(np.pi / 4), np.cos(np.pi / 4), 0],
                    [0, 0, 1],
                ]
            )
            aut = np.concatenate([AUT_D4h, [C8 @ g for g in AUT_D4h]])
        elif symbol in ["32", "3m", "-3m", "-6m2"]:
            # Aut: | 6/mmm / 2/m | = 6
            aut = AUT_D6h
        elif symbol in ["622", "6mm", "6/mmm"]:
            # Aut: 12 22
            C12_hex = np.array(
                [
                    [2 / np.sqrt(3), -1 / np.sqrt(3), 0],
                    [1 / np.sqrt(3), 1 / np.sqrt(3), 0],
                    [0, 0, 1],
                ]
            )
            aut = np.concatenate([AUT_D6h, [C12_hex @ g for g in AUT_D6h]])
        elif symbol in ["1", "-1", "2", "m", "2/m", "4", "-4", "4/m", "3", "-3", "6", "-6", "6/m"]:
            aut = np.array(get_pointgroup_representative_from_symbol("1"))
        else:
            raise ValueError("Unreachable!")

        group = get_pointgroup_representative_from_symbol(symbol)
        permutations = [get_automorphism_permutation(group, g) for g in aut]
        unique_perms = {tuple(p) for p in permutations}
        assert len(permutations) == len(unique_perms)

        all_datum[symbol] = [(g.tolist(), perm) for g, perm in zip(aut, permutations)]

    path = Path(__file__).parent.parent / "src" / "spinspg" / "automorphisms.json"
    with open(path, "w") as f:
        json.dump(all_datum, f)


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from spinspg.spin import SpinOnlyGroupType, get_spin_only_group, solve_procrustes


@pytest.fixture
def nonmagnetic():
    magmoms_nonmagnetic = np.array(
        [
            [0, 0, 0],
            [0, 0, 0],
        ],
        dtype=np.float_,
    )
    return magmoms_nonmagnetic


@pytest.fixture
def collinear():
    magmoms_collinear = np.array(
        [
            [0, 0, 1],
            [0, 0, -2],
        ],
        dtype=np.float_,
    )
    return magmoms_collinear


@pytest.fixture
def coplanar():
    magmoms_coplanar = np.array(
        [
            [0, 0, 1],
            [0, 1, 0],
        ],
        dtype=np.float_,
    )
    return magmoms_coplanar


@pytest.fixture
def noncoplanar():
    magmoms_noncoplanar = np.array(
        [
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ],
        dtype=np.float_,
    )
    return magmoms_noncoplanar


@pytest.mark.parametrize(
    "magmoms_type,sog_type,axis",
    [
        ("nonmagnetic", SpinOnlyGroupType.NONMAGNETIC, None),
        ("collinear", SpinOnlyGroupType.COLLINEAR, [0, 0, 1]),
        ("coplanar", SpinOnlyGroupType.COPLANAR, [1, 0, 0]),
        ("noncoplanar", SpinOnlyGroupType.NONCOPLANAR, None),
    ],
)
def test_spin_only_group(request, magmoms_type, sog_type, axis):
    magmoms = request.getfixturevalue(magmoms_type)
    sog = get_spin_only_group(magmoms, mag_symprec=1e-5)
    assert sog.spin_only_group_type == sog_type
    if axis is not None:
        assert np.allclose(np.cross(sog.axis, axis), 0)

    # Test __str__
    assert str(sog)


FOURFOLD_Z = [
    [0, -1, 0],
    [1, 0, 0],
    [0, 0, 1],
]
MIRROR_X = [
    [-1, 0, 0],
    [0, 1, 0],
    [0, 0, 1],
]


@pytest.mark.parametrize(
    "magmoms_type,linear,expect",
    [
        ("nonmagnetic", FOURFOLD_Z, True),
        ("nonmagnetic", MIRROR_X, True),
        ("collinear", FOURFOLD_Z, True),
        ("collinear", MIRROR_X, True),
        ("coplanar", FOURFOLD_Z, False),
        ("coplanar", MIRROR_X, True),
        ("noncoplanar", FOURFOLD_Z, False),
        ("noncoplanar", MIRROR_X, False),
    ],
)
def test_contain(request, magmoms_type, linear, expect):
    magmoms = request.getfixturevalue(magmoms_type)
    sog = get_spin_only_group(magmoms, mag_symprec=1e-5)
    assert sog.contain(linear) == expect


def test_solve_procrustes():
    A = np.array(
        [
            [1, 0, 0],
            [1, 1, 0],
            [1, 1, 1],
            [0, 0, 1],
        ]
    )
    # Random orthogonal matrix:
    #   a = np.random.random((3, 3))
    #   R = sp.linalg.polar(a)[0]
    R = np.array(
        [
            [-0.03831202, 0.99710779, 0.06563723],
            [0.28470653, -0.05207084, 0.95719947],
            [0.95784883, 0.05535959, -0.28188816],
        ]
    )
    B = A @ R.T

    R_actual = solve_procrustes(A, B)
    assert np.allclose(R_actual, R)



import numpy as np

from spinspg import get_spin_symmetry

# Antiferromagnetic rutile structure
a = 4.87
c = 3.31
x_4f = 0.695169
lattice = np.diag([a, a, c])
positions = np.array(
    [  # Fractional coordinates
        [0, 0, 0],  # Mn(2a)
        [0.5, 0.5, 0.5],  # Mn(2a)
        [x_4f, x_4f, 0],  # F(4f)
        [-x_4f, -x_4f, 0],  # F(4f)
        [-x_4f + 0.5, x_4f + 0.5, 0.5],  # F(4f)
        [x_4f + 0.5, -x_4f + 0.5, 0.5],  # F(4f)
    ]
)
numbers = np.array([0, 0, 1, 1, 1, 1])
magmoms = np.array(
    [  # In Cartesian coordinates
        [0, 0, 2.5],
        [0, 0, -2.5],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
    ]
)

# Find spin symmetry operations
sog, rotations, translations, spin_rotations = get_spin_symmetry(
    lattice, positions, numbers, magmoms
)

print(f"Spin-only group: {sog}")  # COLLINEAR(axis=[0. 0. 1.])

# Some operations have nontrivial spin rotations
idx = 2
print(f"Rotation ({idx})\n{rotations[idx]}")
print(f"Translation ({idx})\n{translations[idx]}")
print(f"Spin rotation ({idx})\n{spin_rotations[idx]}")  # -> diag([1, 1, -1])



"""Group for spin symmetry operations."""
from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from hsnf import column_style_hermite_normal_form
from spglib import get_symmetry_dataset

from spinspg.permutation import Permutation, get_symmetry_permutations
from spinspg.spin import SpinOnlyGroup, get_spin_only_group, solve_procrustes
from spinspg.utils import (
    NDArrayFloat,
    NDArrayInt,
    is_integer_array,
    ndarray2d_to_integer_tuple,
)


@dataclass
class NonmagneticSymmetry:
    """Crystal structure with symmetry operations.

    Attributes
    ----------
    prim_lattice: array, (3, 3)
        Primitive basis vectors w.r.t. nonmagnetic symmetry
    prim_rotations: array[int], (order, 3, 3)
        w.r.t. ``prim_lattice``
    prim_translations: array, (order, 3)
        w.r.t. ``prim_lattice``
    prim_permutations: array[int], (order, num_sites)
        ``num_sites`` is a number of sites in input cell.
        ``(prim_rotations[p], prim_translations[p])`` moves the ``i``-th site to the ``prim_permutations[p][i]``
    prim_centerings: array[int], (nc, 3)
        Centering translations w.r.t. ``prim_lattice``
    prim_centering_permutations: (nc, N)
    transformation: array[int], (3, 3)
        Transformation matrix from primitive to given cell
    """

    prim_lattice: NDArrayFloat
    prim_rotations: NDArrayInt
    prim_translations: NDArrayFloat
    prim_permutations: list[Permutation]
    prim_centerings: NDArrayInt
    prim_centering_permutations: list[Permutation]
    transformation: NDArrayInt


def get_symmetry_with_cell(
    lattice: NDArrayFloat,
    positions: NDArrayFloat,
    numbers: NDArrayInt,
    symprec: float,
    angle_tolerance: float,
) -> NonmagneticSymmetry:
    """Find spatial symmetry operations from nonmagnetic crystal structure."""
    dataset = get_symmetry_dataset((lattice, positions, numbers), symprec, angle_tolerance)
    rotations = dataset["rotations"]
    translations = dataset["translations"]
    prim_lattice = dataset["primitive_lattice"]

    # Unique by rotation parts
    uniq_rotations = []
    uniq_translations = []
    centerings = []
    found_rotations = set()
    for rot, trans in zip(rotations, translations):
        if np.allclose(rot, np.eye(3)):
            centerings.append(trans)

        rot_int = ndarray2d_to_integer_tuple(rot)
        if rot_int in found_rotations:
            continue
        uniq_rotations.append(rot)
        uniq_translations.append(trans)
        found_rotations.add(rot_int)

    # Primitive transformation
    tmat = np.linalg.inv(prim_lattice.T) @ lattice.T
    assert np.isclose(np.abs(np.linalg.det(tmat)), len(centerings))

    # Permutations of sites
    prim_permutations = get_symmetry_permutations(
        lattice,
        positions,
        numbers,
        rotations=uniq_rotations,
        translations=uniq_translations,
        symprec=symprec,
    )
    prim_centering_permutations = get_symmetry_permutations(
        lattice,
        positions,
        numbers,
        rotations=[np.eye(3) for _ in range(len(centerings))],
        translations=centerings,
        symprec=symprec,
    )

    # To primitive basis (never take modulus!)
    prim_rotations = []
    prim_translations = []
    invtmat = np.linalg.inv(tmat)
    for rot, trans in zip(uniq_rotations, uniq_translations):
        prim_rotations.append(np.around(tmat @ rot @ invtmat).astype(np.int_))
        prim_translations.append(tmat @ trans)

    prim_centerings = []
    for centering in centerings:
        prim_centering = tmat @ centering
        assert is_integer_array(prim_centering)
        prim_centerings.append(np.around(prim_centering).astype(int))

    return NonmagneticSymmetry(
        prim_lattice=prim_lattice,
        prim_rotations=np.array(prim_rotations),
        prim_translations=np.array(prim_translations),
        prim_permutations=prim_permutations,
        prim_centerings=np.array(prim_centerings),
        prim_centering_permutations=prim_centering_permutations,
        transformation=np.around(tmat).astype(np.int_),
    )


@dataclass
class SpinSymmetryOperation:
    """Spin symmetry operation.

    Attributes
    ----------
    rotations: array[int], (3, 3)
    translation: array, (3, )
    spin_rotation: array, (3, 3)
    """

    rotation: NDArrayInt
    translation: NDArrayFloat
    spin_rotation: NDArrayFloat


@dataclass
class SpinSpaceGroup:
    """Spin space group.

    Attributes
    ----------
    primitive_lattice: array, (3, 3)
        primitive_lattice[i] is the i-th primitive basis vector for maximal space subgroup
    spin_only_group: SpinOnlyGroup
    spin_translation_coset: list[SpinSymmetryOperation]
        N.B. translation parts are distinct
    centerings: array, (?, 3)
    nontrivial_coset: list[SpinSymmetryOperation]
        N.B. rotation parts are distinct
    transformation: array[int], (3, 3)
        Transformation matrix from primitive to given cell
    """

    prim_lattice: NDArrayFloat
    spin_only_group: SpinOnlyGroup
    spin_translation_coset: list[SpinSymmetryOperation]
    prim_centerings: NDArrayFloat
    nontrivial_coset: list[SpinSymmetryOperation]
    transformation: NDArrayInt


def get_primitive_spin_symmetry(
    nonmagnetic_symmetry: NonmagneticSymmetry, magmoms: NDArrayFloat, mag_symprec: float
) -> SpinSpaceGroup:
    """Return spin space group symmetry.

    Parameters
    ----------
    nonmagnetic_symmetry : NonmagneticSymmetry
    magmoms : array, (num_sites, 3)
    mag_symprec : float

    Returns
    -------
    SpinSpaceGroup
    """
    # Centerings for maximal space subgroup of spin space group
    stg_centerings = []
    stg_centering_permutations = []
    for centering, perm in zip(
        nonmagnetic_symmetry.prim_centerings, nonmagnetic_symmetry.prim_centering_permutations
    ):
        if np.max(np.linalg.norm(magmoms[perm.permutation] - magmoms, axis=1)) < mag_symprec:
            stg_centerings.append(centering)
            stg_centering_permutations.append(perm)
    assert len(nonmagnetic_symmetry.prim_centerings) % len(stg_centerings) == 0

    # Transformation matrix to primitive cell of maximal space subgroup
    stg_vectors = np.concatenate(
        [
            nonmagnetic_symmetry.transformation,  # (3, 3)
            np.array(stg_centerings).T,  # (3, ?)
        ],
        axis=1,
    )
    tmat_stg, _ = column_style_hermite_normal_form(stg_vectors)
    tmat_stg = tmat_stg[:, :3]  # (3, 3)
    invtmat_stg = np.linalg.inv(tmat_stg)

    # Spin only group
    spin_only_group = get_spin_only_group(magmoms, mag_symprec)

    prim_centerings = [invtmat_stg @ centering for centering in stg_centerings]
    # prim_spin_lattice.T @ transformation == nonmagnetic_symmetry.prim_lattice.T @ nonmagnetic_symmetry.transformation
    transformation = invtmat_stg @ nonmagnetic_symmetry.transformation
    assert is_integer_array(transformation)
    transformation = np.around(transformation).astype(np.int_)

    # Spin translation group search
    spin_translation_coset = []
    found_stg_centerings = []  # type: ignore
    for centering, perm in zip(
        nonmagnetic_symmetry.prim_centerings, nonmagnetic_symmetry.prim_centering_permutations
    ):
        # Two centerings are equivalent in primitive cell of spin translation group if they
        # are translated to each other by lattice translations in `transformation`.
        is_new_centering = True
        for other in found_stg_centerings:
            residual = invtmat_stg @ (centering - other)
            residual -= np.rint(residual)
            if np.allclose(residual, 0):
                is_new_centering = False
                break
        if not is_new_centering:
            continue
        found_stg_centerings.append(centering)

        # Search W in O(3) s.t. new_magmoms @ W.T = magmoms[perm.permutation]
        new_magmoms = magmoms.copy()
        perm_magmoms = magmoms[perm.permutation]
        W = solve_procrustes(new_magmoms, perm_magmoms)
        if spin_only_group.contain(W):
            # Chose W as identify if W belongs to the spin only group
            W = np.eye(3, dtype=np.float_)

        new_magmoms = new_magmoms @ W.T
        if np.max(np.linalg.norm(new_magmoms - perm_magmoms, axis=1)) < mag_symprec:
            # w.r.t. primitive cell of spin space group
            reduced_centering = invtmat_stg @ centering
            spin_translation_coset.append(
                SpinSymmetryOperation(
                    rotation=np.eye(3, dtype=np.int_),
                    translation=reduced_centering,
                    spin_rotation=W,
                )
            )

    assert len(nonmagnetic_symmetry.prim_centerings) % len(found_stg_centerings) == 0

    # Spin space group search
    nontrivial_coset = []
    for rot, trans, perm in zip(
        nonmagnetic_symmetry.prim_rotations,
        nonmagnetic_symmetry.prim_translations,
        nonmagnetic_symmetry.prim_permutations,
    ):
        # Point group symmetry compatible with the primitive cell
        rot_prim = invtmat_stg @ rot @ tmat_stg
        if not is_integer_array(rot_prim):
            continue

        # Need to consider centerings for subgroup
        for centering, centering_perm in zip(
            nonmagnetic_symmetry.prim_centerings, nonmagnetic_symmetry.prim_centering_permutations
        ):
            new_perm: Permutation = centering_perm * perm
            new_magmoms = magmoms.copy()
            perm_magmoms = magmoms[new_perm.permutation]
            W = solve_procrustes(new_magmoms, perm_magmoms)
            if spin_only_group.contain(W):
                # Chose W as identify if W belongs to the spin only group
                W = np.eye(3, dtype=np.float_)

            new_magmoms = new_magmoms @ W.T
            if np.max(np.linalg.norm(new_magmoms - perm_magmoms, axis=1)) < mag_symprec:
                # w.r.t. primitive cell of spin space group
                new_trans = centering + trans
                nontrivial_coset.append(
                    SpinSymmetryOperation(
                        rotation=rot_prim,
                        translation=invtmat_stg @ new_trans,
                        spin_rotation=W,
                    )
                )
                break

    # Transform centerings to primitive cell of spin space group
    prim_spin_lattice = tmat_stg.T @ nonmagnetic_symmetry.prim_lattice

    return SpinSpaceGroup(
        prim_lattice=prim_spin_lattice,
        spin_only_group=spin_only_group,
        spin_translation_coset=spin_translation_coset,
        prim_centerings=prim_centerings,
        nontrivial_coset=nontrivial_coset,
        transformation=transformation,
    )



import numpy as np
import pytest
from spglib import get_magnetic_symmetry

from spinspg.core import get_spin_symmetry
from spinspg.group import get_primitive_spin_symmetry, get_symmetry_with_cell
from spinspg.spin import SpinOnlyGroupType


def test_get_symmetry_with_cell(fcc):
    lattice, positions, numbers, _ = fcc
    symmetry = get_symmetry_with_cell(lattice, positions, numbers, 1e-5, -1)
    assert symmetry.prim_rotations.shape == (48, 3, 3)
    assert symmetry.prim_centerings.shape == (4, 3)


def test_spin_space_group_fcc(fcc):
    lattice, positions, numbers, magmoms = fcc
    symprec = 1e-5
    ns = get_symmetry_with_cell(lattice, positions, numbers, symprec, -1)
    ssg = get_primitive_spin_symmetry(ns, magmoms, symprec)

    assert ssg.spin_only_group.spin_only_group_type == SpinOnlyGroupType.COPLANAR
    assert len(ssg.spin_translation_coset) == 2


def test_spin_space_group_kagome(layer_triangular_kagome):
    lattice, positions, numbers, magmoms = layer_triangular_kagome
    symprec = 1e-5
    ns = get_symmetry_with_cell(lattice, positions, numbers, symprec, -1)
    ssg = get_primitive_spin_symmetry(ns, magmoms, symprec)

    assert ssg.spin_only_group.spin_only_group_type == SpinOnlyGroupType.COPLANAR
    assert len(ssg.spin_translation_coset) == 1

    kernel_pointgroup = []
    for ops in ssg.nontrivial_coset:
        if np.allclose(ops.spin_rotation, np.eye(3)):
            kernel_pointgroup.append(ops.rotation)
    assert len(kernel_pointgroup) == 4  # 2/m
    assert len(ssg.nontrivial_coset) == 24  # 6/mmm


def test_spin_space_group_rutile(rutile):
    lattice, positions, numbers, magmoms = rutile
    symprec = 1e-5
    ns = get_symmetry_with_cell(lattice, positions, numbers, symprec, -1)
    ssg = get_primitive_spin_symmetry(ns, magmoms, symprec)

    assert ssg.spin_only_group.spin_only_group_type == SpinOnlyGroupType.COLLINEAR
    assert np.allclose(np.cross(ssg.spin_only_group.axis, [0, 0, 1]), 0)
    assert len(ssg.spin_translation_coset) == 1
    assert len(ssg.prim_centerings) == 1
    assert np.isclose(np.linalg.det(ssg.transformation), 1)

    kernel_pointgroup = []
    for ops in ssg.nontrivial_coset:
        if np.allclose(ops.spin_rotation, np.eye(3)):
            kernel_pointgroup.append(ops.rotation)

    assert len(ssg.nontrivial_coset) == 16  # 4/mmm
    assert len(kernel_pointgroup) == 8  # 4/m


@pytest.mark.parametrize(
    "testcase,spin_only_group_type,axis",
    [
        ("Cr_in_Cr2O3", SpinOnlyGroupType.COLLINEAR, [1, 0, 0]),
        ("Pr_in_PrScSb", SpinOnlyGroupType.COLLINEAR, [0, 0, 1]),
        ("Mn_in_Mn3ReO6", SpinOnlyGroupType.COPLANAR, [0, 0, 1]),
        ("Ni_in_NiTa2O6", SpinOnlyGroupType.COLLINEAR, [1, -1, 0]),
    ],
)
def test_spin_space_groups(request, testcase, spin_only_group_type, axis):
    lattice, positions, numbers, magmoms = request.getfixturevalue(testcase)
    symprec = 1e-5
    ns = get_symmetry_with_cell(lattice, positions, numbers, symprec, -1)
    ssg = get_primitive_spin_symmetry(ns, magmoms, symprec)

    assert ssg.spin_only_group.spin_only_group_type == spin_only_group_type
    assert np.allclose(np.cross(ssg.spin_only_group.axis, axis), 0)

    num_sym = (
        len(ssg.nontrivial_coset)
        * len(ssg.spin_translation_coset)
        * np.around(np.abs(np.linalg.det(ssg.transformation))).astype(int)
    )
    mag_symmetry = get_magnetic_symmetry((lattice, positions, numbers, magmoms), symprec=symprec)
    assert num_sym >= len(mag_symmetry["rotations"])


def test_get_spin_symmetry(rutile):
    lattice, positions, numbers, magmoms = rutile
    sog, rotations, translations, spin_rotations = get_spin_symmetry(
        lattice, positions, numbers, magmoms
    )

    assert sog.spin_only_group_type == SpinOnlyGroupType.COLLINEAR
    assert np.allclose(np.cross(sog.axis, [0, 0, 1]), 0)  # parallel to [0, 0, 1]

    identity = np.eye(3)
    mz = np.diag([1, 1, -1])
    expects = [
        (
            # (1) x, y, z; 1
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (2) -x, -y, z; 1
            [[-1, 0, 0], [0, -1, 0], [0, 0, 1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (3) -y + 1/2, x + 1/2, z + 1/2; mz
            [[0, -1, 0], [1, 0, 0], [0, 0, 1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (4) y + 1/2, -x + 1/2, z + 1/2; mz
            [[0, 1, 0], [-1, 0, 0], [0, 0, 1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (5) -x + 1/2, y + 1/2, -z + 1/2; mz
            [[-1, 0, 0], [0, 1, 0], [0, 0, -1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (6) x + 1/2, -y + 1/2, -z + 1/2; mz
            [[1, 0, 0], [0, -1, 0], [0, 0, -1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (7) y, x, -z; 1
            [[0, 1, 0], [1, 0, 0], [0, 0, -1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (8) -y, -x, -z; 1
            [[0, -1, 0], [-1, 0, 0], [0, 0, -1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (9) -x, -y, -z; 1
            [[-1, 0, 0], [0, -1, 0], [0, 0, -1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (10) x, y, -z; 1
            [[1, 0, 0], [0, 1, 0], [0, 0, -1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (11) y + 1/2, -x + 1/2, -z + 1/2; mz
            [[0, 1, 0], [-1, 0, 0], [0, 0, -1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (12) -y + 1/2, x + 1/2, -z + 1/2; mz
            [[0, -1, 0], [1, 0, 0], [0, 0, -1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (13) x + 1/2, -y + 1/2, z + 1/2; mz
            [[1, 0, 0], [0, -1, 0], [0, 0, 1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (14) -x + 1/2, y + 1/2, z + 1/2; mz
            [[-1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [0.5, 0.5, 0.5],
            mz,
        ),
        (
            # (15) -y, -x, z; 1
            [[0, -1, 0], [-1, 0, 0], [0, 0, 1]],
            [0, 0, 0],
            identity,
        ),
        (
            # (16) y, x, z; 1
            [[0, 1, 0], [1, 0, 0], [0, 0, 1]],
            [0, 0, 0],
            identity,
        ),
    ]

    found = [False for _ in expects]
    assert len(rotations) == len(expects)
    for rot, trans, srot in zip(rotations, translations, spin_rotations):
        for i in range(len(expects)):
            if (not found[i]) and np.allclose(rot, expects[i][0]):
                assert np.allclose(trans, expects[i][1])
                assert np.allclose(srot, expects[i][2])
                found[i] = True

    assert all(found)



"""Import top APIs and version."""
from importlib.metadata import PackageNotFoundError, version

from spinspg.core import get_spin_symmetry  # noqa: F401

# https://github.com/pypa/setuptools_scm/#retrieving-package-version-at-runtime
try:
    __version__ = version("spinspg")
except PackageNotFoundError:
    # package is not installed
    pass
